//---------------------------------------------------------------------------------------------
//	COMPLEX DATATYPE USED FOR THE MATRIX DEFINITION
//---------------------------------------------------------------------------------------------

	//---------------------------------------------------------------------------------------------
	//	STANDARD PRECISION COMPLEX
	//---------------------------------------------------------------------------------------------

		class Complex {
		double re;
		double im;

		public:
		  
		friend class ArbFloat;
		friend class ArbComplex;

		  
		/*!
		The complex number is initialised with zero real and imaginary
		parts by default.
		*/
		Complex(double a = 0, double b = 0);
		
		Complex(const Complex& a);
		~Complex();

		Complex& operator=(const Complex& a);
		Complex& operator=(const double a);
		Complex& operator=(const double* a);
 		Complex& operator=(const int a);

		friend Complex operator/(Complex &a, double &b)
		{
			return Complex(a.re/b,a.im/b);
		}
		
		friend Complex operator/(Complex &a, int &b)
		{
			return Complex(a.re/(double)b,a.im/(double)b);
		}
		
		//! Get the real part
		/*! \return the real part */
		double real() const { return re; }
		double reald() const { return re; }
		//! Get the imaginary part
		/*! \return the imaginary part */
		double imag() const { return im; }
		double imagd() const { return im; }

		void addreal( double x) { re+=x; }
		void addimag( double x) { im+=x;}

		void zero()
		{ re=0; im=0; }

		//! Assign to the real part
		/*! \param r the real part */
		void real(double r) { re = r; }
		//! Assign to the imaginary part
		/*! \param i the imaginary part */
		void imag(double i) { im = i; }

		//! Returns the square norm of the complex number.
		double norm() const;
		//! Returns the norm (modulus) of the complex number.  
		double abs() const;

		char* reals( char* in) const { sprintf(in, "%+.20e", re); return in; }
		char* imags( char* in) const { sprintf(in, "%+.20e", im); return in; }
		
		
		//! Overloaded complex sum 
		Complex& operator+=(const Complex& a);
		//! Overloaded real sum
		Complex& operator+=(double a);

		//! Overloaded complex subtraction
		Complex& operator-=(const Complex& a);
		//! Overloaded real subtraction
		Complex& operator-=(double a);

		//! Overloaded complex multiplication
		Complex& operator*=(const Complex& a);
		//! Overloaded real multiplication
		Complex& operator*=(double a);
		
		//! Overloaded complex division
		Complex& operator/=(const Complex& a);
		//! Overloaded real division
		Complex& operator/=(double a);

		int operator !=(const double &b) const;
		int operator !=(const Complex& a) const;
		
		void Conj()
		{
		  im*=-1.;
		}
		
		// These functions are specified as friends, but
		// they do not actually reference private member

		//! Complex conjugation
		friend  Complex conj(const Complex& c);

		//! Complex addition
		friend  Complex operator + (const Complex& c1, const Complex& c2)
		{  return Complex(c1) += c2; }

		//! Real addition 
		friend  Complex operator + (const Complex& c, double f)
		{ return Complex(c) += f; }
		
		//! Real addition 
		friend  Complex operator + (double f, const Complex& c)
		{ return Complex(c) += f; }
		
		//! Complex subtraction 
		friend  Complex operator - (const Complex& c1, const Complex& c2)
		{  return Complex(c1) -= c2; }

		//! Real subtraction  
		friend  Complex operator - (const Complex& c, double f)
		{  return Complex(c) -= f; }
		
		//! Real subtraction  
		friend  Complex operator - (double f, const Complex& c)
		{  return Complex(f) -= c; }

		//! Complex multiplication
		friend  Complex operator * (const Complex& c1, const Complex& c2)
		{  return Complex(c1) *= c2; }

		//! Real multiplication
		friend  Complex operator * (const Complex& c, double f)
		{  return Complex(c) *= f; }

		//! Real multiplication 
		friend  Complex operator * (double f, const Complex& c)
		{  return Complex(c) *= f; }

		//! Complex division
		friend  Complex operator / (const Complex& c1, const Complex& c2)
		{  return Complex(c1) /= c2; }

		//! Real division 
		friend  Complex operator / (const Complex& c, double f)
		{  return Complex(c) /= f; }
		
		//! Real division 
		friend  Complex operator / (double f, const Complex& c)
		{  return Complex(f) /= c; }

		//! Unary negation 
		friend  Complex operator - (const Complex& c);

		// More friends for the purpose of injections. e.g.
		// real(c1+c2) instead of tmp = c1+c2; tmp.real();

		//! Real part
		friend double real(const Complex& c) { return c.real(); }
		//! Imaginary part
		friend double imag(const Complex& c) { return c.imag(); }
		//! Square norm
		friend double norm(const Complex& c) { return c.norm(); }
		//! Complex norm (modulus)
		friend double abs(const Complex& c) { return c.abs(); }

		void Polar(double abs, double arg)
		{
		
			re=abs*cos(arg);
			im=abs*sin(arg);
			
		}
		
		void Polar(double *abs, double *arg)
		{
			abs[0]=sqrt(re*re+im*im);
			arg[0]=atan(im/re);

		}
		
		};
		
		int Complex::operator !=(const double &b) const
		{
		
			if(im!=0) return 1;
			if(re==0) { return 0; } else { return 1; }
			
		}

		int Complex::operator !=(const Complex& a) const
		{
			
			if(this->im!=a.im || this->re != a.re){
				return 1;
			}else{
				return 0;
			}
			
		}


		Complex::Complex(double a, double b)
		: re(a), im(b){}

		Complex::Complex(const Complex& a)
		: re(a.re), im(a.im){}

		Complex::~Complex()
		{}

		Complex& Complex::operator=(const Complex& a)
		{ re = a.re; im = a.im; return *this; }

		Complex& Complex::operator=(const double* a)
		{ re = a[0]; im = a[1]; return *this; }
		
		Complex& Complex::operator=(const double a)
		{ re = a; im = 0; return *this; }

		Complex& Complex::operator=(const int a)
		{ re = a; im = 0; return *this; }

/*		Complex& Complex::operator=(const int a)
		{ re = a; im = 0; return *this; }*/
		
		double Complex::abs() const { 
		return sqrt(norm()); 

		}

		double Complex::norm() const
		{  return re*re + im*im; }
		
		Complex& Complex::operator+=(const Complex& a)
		{  re += a.re;
		im += a.im;
		return *this;
		}
		
		Complex& Complex::operator+=(double a)
		{  re += a;
		return *this;
		}

		Complex& Complex::operator-=(const Complex& a)
		{  re -= a.re;
		im -= a.im;
		return *this;
		}
		
		Complex& Complex::operator-=(double a)
		{  re -= a;
		return *this;
		}

		Complex& Complex::operator*=(const Complex& a)
		{ 
		double sre = re * a.re - im * a.im;
		im = re * a.im + im * a.re;
		re = sre;
		return *this;
		}
		
		Complex& Complex::operator*=(double a)
		{  re *= a;
		im *= a;
		return *this;
		}
		
		Complex& Complex::operator/=(const Complex& a)
		{
		double norm2_a_1 = 1.0/(a.re * a.re + a.im * a.im);
		double sre = (re * a.re + im * a.im) * norm2_a_1;
		im = (im * a.re - re * a.im) * norm2_a_1;
		re = sre;
		return *this;
		}

		Complex& Complex::operator/=(double a)
		{  re /= a;
		im /= a;
		return *this;
		}

		Complex conj(const Complex& c)
		{  return Complex(c.re, -c.im); }

		Complex operator - (const Complex& c)
		{  return Complex(-c.re, -c.im); }




